grammar Javali; // parser grammar, parses streams of tokens

@header {
	// Java header
	package cd.frontend.parser;
}



// PARSER RULES
literal
    :   'null'  # NullLiteral
    |   Boolean # BoolLiteral
    |   Integer # IntLiteral
    ;

// Types
type
    : primitiveType
    | referenceType
    ;

referenceType
    : Ident
    | arrayType
    ;

arrayType
    : Ident '[' ']'
    | primitiveType '[' ']'
    ;

primitiveType
    :   'boolean'
    |   'int'
    ;

// Program structure
unit
    : classDecl+ EOF
	;

classDecl
    : 'class' Ident ('extends' Ident)? '{' memberList '}'
    ;

memberList
    : (varDecl | methodDecl)*
    ;

varDecl
    : type Ident (',' Ident)* ';'
    ;

methodDecl
    : (type | 'void') Ident '(' formalParamList? ')' '{' varDecl* stmt* '}'
    ;

formalParamList
    : type Ident (',' type Ident)*
    ;

// Statements
stmt
    : assignmentStmt
    | methodCallStmt
    | ifStmt
    | whileStmt
    | returnStmt
    | writeStmt
    ;

stmtBlock
    : '{' stmt* '}'
    ;

methodCallStmt
    : Ident '(' actualParamList? ')' ';'                 # LocalMethodCallStmt
    | identAccess '.' Ident '(' actualParamList? ')' ';' # ObjectMethodCallStmt
    ;

assignmentStmt
    : identAccess '=' (expr | newExpr | readExpr) ';'
    ;

writeStmt
    : 'write' '(' expr ')' ';' # Write
    | 'writeln' '(' ')' ';'    # WriteLn
    ;

ifStmt
    : 'if' '(' expr ')' stmtBlock ('else' stmtBlock)?
    ;

whileStmt
    : 'while' '(' expr ')' stmtBlock
    ;

returnStmt
    : 'return' expr? ';'
    ;

// Expressions
newExpr
    : 'new' Ident '(' ')'               # NewObject
    | 'new' Ident '[' expr ']'          # NewObjectArray
    | 'new' primitiveType '[' expr ']'  # NewPrimitiveArray
    ;

readExpr
    : 'read' '(' ')'
    ;

actualParamList
    : expr (',' expr)*
    ;

identAccess
    : Ident                                         # AccessLocalField
    | 'this'                                        # AccessThis
    | identAccess '.' Ident                         # AccessObjectField
    | identAccess '[' expr ']'                      # AccessArray
    | Ident '(' actualParamList? ')'                # AccessLocalMethod
    | identAccess '.' Ident '(' actualParamList? ')'# AccessObjectMethod
    ;

expr
    : literal                       # ExprConstant
    | identAccess                   # ExprIdentAccess
    | '(' expr ')'                  # ExprParentheses
    | ('+'|'-'|'!') expr            # ExprUnary
    | '(' referenceType ')' expr    # ExprCast
    | expr ('*'|'/'|'%') expr       # ExprBinary
    | expr ('+'|'-') expr           # ExprBinary
    | expr ('<'|'>'|'<='|'>=') expr # ExprBinary
    | expr ('=='|'!=') expr         # ExprBinary
    | expr '&&' expr                # ExprBinary
    | expr '||' expr                # ExprBinary
    ;


// LEXER RULES
fragment
Letter
	:	'A'..'Z'
	|	'a'..'z'
	;

fragment
Digit
    :   '0'..'9'
    ;

fragment
HexDigit
    :   Digit
    |   'a'..'f'
    |   'A'..'F'
    ;

fragment
Decimal
    :   '0'
    |   '1'..'9' Digit*
    ;

fragment
Hexadecimal
    :   ('0x'|'0X') HexDigit+
    ;

Integer
    :   Decimal
    |   Hexadecimal
    ;

Boolean
    :   'false'
    |   'true'
    ;

Ident
    :   Letter (Letter|Digit)*
    ;



// comments and white space does not produce tokens:
COMMENT
	:	'/*' .*? '*/' -> skip
	;

LINE_COMMENT
	:	'//' ~('\n'|'\r')* -> skip
	;

WS
	:	(' '|'\r'|'\t'|'\n') -> skip
	;


// handle characters which failed to match any other token
ErrorCharacter : . ;
